#include "config.h"
#include "dosbox.h"

#if defined (WIN32)

#include "serialport/porttalk.h"
#include "inout.h"

static Bit16s hardopldiff;

static void write_hwio(Bitu port,Bitu val,Bitu /*iolen*/) {
	if(port<0x388) port += hardopldiff;
	//LOG_MSG("write port %x",port);
	outportb(port, val);
}

static Bitu read_hwio(Bitu port,Bitu /*iolen*/) {
	if(port<0x388) port += hardopldiff;
	//LOG_MSG("read port %x",port);
	Bitu retval = inportb(port);
	return retval;
}

bool hwopl_dirty=false;

static IO_ReadHandleObject* hwOPL_ReadHandler[10] ;
static IO_WriteHandleObject* hwOPL_WriteHandler[10];

const Bit16u oplports[]={
		0x0,0x1,0x2,0x3,0x8,0x9,
		0x388,0x389,0x38A,0x38B};

void HARDOPL_Init(Bitu hardwareaddr, Bitu blasteraddr)
{
	//LOG_MSG("hoplinit");
	if(!(hardwareaddr==0x210 || hardwareaddr==0x220 || hardwareaddr==0x230 ||
		 hardwareaddr==0x240 || hardwareaddr==0x250 || hardwareaddr==0x260 ||
		 hardwareaddr==0x280)) {
			LOG_MSG("Invalid hardware base address. Aborting.");
			return;
		 }
	
	if(!initPorttalk())
	{
		LOG_MSG("Porttalk not loaded. Aborting.");
		return;
	}


	hardopldiff=hardwareaddr-blasteraddr;
	hwopl_dirty=true;

	// map the port
	LOG_MSG("Port mappings hardware -> DOSBox:");
	for(int i = 0; i < 10; i++)
	{
		hwOPL_ReadHandler[i]=new IO_ReadHandleObject();
		hwOPL_WriteHandler[i]=new IO_WriteHandleObject();
		Bit16u port=oplports[i];
		if(i<6) port+=blasteraddr; 
		hwOPL_ReadHandler[i]->Install(port,read_hwio,IO_MB);
		hwOPL_WriteHandler[i]->Install(port,write_hwio,IO_MB);

		if(i<6) port+=hardopldiff;
		LOG_MSG("%x -> %x",port,i<6?port-hardopldiff:port);
		addIOPermission(port);
	}
	setPermissionList();
}

void HWOPL_Cleanup()
{
	if(hwopl_dirty) {
		for(int i = 0; i < 10; i++) {
			delete hwOPL_ReadHandler[i];
			delete hwOPL_WriteHandler[i];
		}
		hwopl_dirty=false;
	}

}

#endif
