/*
 *  Copyright (C) 2002-2006  The DOSBox Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <string.h>
#include <ctype.h>

#include "dosbox.h"

#include "support.h"
#include "inout.h"
#include "pic.h"
#include "setup.h"
#include "timer.h"
#include "bios.h"					// SetLPTPorts(..)

#include "parport.h"
#include "directlpt_win32.h"
#include "misc_util.h"
#include "dos_inc.h"

// LPT1 - LPT3 objects
static CParallel *parallel1 = 0;
static CParallel *parallel2 = 0;
static CParallel *parallel3 = 0;


bool device_LPT::Read(Bit8u * data,Bit16u * size) 
{
	*size=0;
	LOG(LOG_DOSMISC,LOG_NORMAL)("LPTDEVICE:Read called");
	return true;
}


bool device_LPT::Write(Bit8u * data,Bit16u * size) 
{
	for (Bit16u i=0; i<*size; i++)
	{
		if(!pportclass->Putchar(data[i])) return false;
	}

	return true;
}

bool device_LPT::Seek(Bit32u * pos,Bit32u type) 
{
	*pos = 0;
	return true;
}

bool device_LPT::Close() 
{
	return false;
}

Bit16u device_LPT::GetInformation(void) 
{
	return 0x80A0;
};
char* lptname[]={"LPT1","LPT2","LPT3"};
device_LPT::device_LPT(Bit8u num, class CParallel* pp)
{
	pportclass = pp;
	SetName(lptname[num-1]);
	//this->name = lptname[num];
	this->num = num;
	
}

device_LPT::~device_LPT()
{
	//LOG_MSG("del");
}


// Usage of FastDelegates (http://www.codeproject.com/cpp/FastDelegate.asp)
// as I/O-Array would make many of these unneccessary. (member function pointer)

//Some defines for repeated functions

#define PARALLEL_UPDATE(number) \
void PAR##number##_Update(void) {	\
	parallel##number->Timer ();	\
}

#define PARALLEL_WRITE_TREE(number) \
static void PAR##number##_Write (Bitu port, Bitu val, Bitu iolen) {	\
	switch (port & 0x7) {	\
	case 0:	\
		parallel##number ->Write_PR (val);	\
		return;	\
	case 1:	\
		parallel##number ->Write_IOSEL (val);	\
		return;	\
	case 2:	\
		parallel##number ->Write_CON (val);	\
		return;	\
	}	\
}

#define PARALLEL_READ_TREE(number) \
static Bitu PAR##number##_Read (Bitu port, Bitu iolen) {	\
	switch (port & 0x3) {	\
	case 0:	\
		return parallel##number ->Read_PR ();	\
	case 1:	\
		return parallel##number ->Read_SR ();	\
	case 2:	\
		return parallel##number ->Read_COM ();	\
	}	\
	return 0;	\
}

//The Functions

PARALLEL_UPDATE(1);
PARALLEL_UPDATE(2);
PARALLEL_UPDATE(3);

PARALLEL_WRITE_TREE(1);
PARALLEL_WRITE_TREE(2);
PARALLEL_WRITE_TREE(3);

PARALLEL_READ_TREE(1);
PARALLEL_READ_TREE(2);
PARALLEL_READ_TREE(3);

#undef PARALLEL_UPDATE
#undef PARALLEL_WRITE_TREE
#undef PARALLEL_READ_TREE

void CParallel::Timer (void) {

	Timer2 ();
}

/*****************************************************************************/
/* Interrupt Enable Register, also MSB of Divisor Latch (r/w)               **/
/*****************************************************************************/
/*
void CParallel::Write_MSR (Bit8u val) {
#ifdef SERIALPORT_DEBUGMSG
	LOG_UART ("Serial port at %x: Write to MSR, value 0x%x", base, val);
#endif
}*/

/*****************************************************************************/
/* Write_reserved                                                           **/
/*****************************************************************************/
void CParallel::Write_reserved (Bit8u data, Bit8u address) {
	LOG_MSG("Serial port at %x: Write to reserved register, value 0x%x, register %x", base, data, address);
}


/*****************************************************************************/
/* Bit Access                                                               **/
/*****************************************************************************/
/*
bool CParallel::getCTS () {
	return true;
}
*//*
void CParallel::setRI (bool value) {
	
}
*/
/*****************************************************************************/
/* Initialisation                                                           **/
/*****************************************************************************/
/*void CSerial::Init_Registers (Bit32u initbps, Bit8u bytesize, 
                              const char *parity, Bit8u stopbits) {
	
}*/

CParallel::CParallel(IO_ReadHandler* rh, IO_WriteHandler* wh,
			TIMER_TickHandler TimerHandler, 
			Bit16u initbase, Bit8u initirq) {
	base = initbase;
	irq = initirq;
	TimerHnd = NULL;
	//TimerHnd = th;				// for destructor
	//TIMER_AddTickHandler(TimerHnd);

	for (Bitu i = 0; i <= 2; i++) {
		WriteHandler[i].Install (i + base, wh, IO_MB);
		ReadHandler[i].Install (i + base, rh, IO_MB);
	}
	mydosdevice=new device_LPT(1, this);
	DOS_AddDevice(mydosdevice);
};

CParallel::~CParallel(void) {
	
	if(TimerHnd) TIMER_DelTickHandler(TimerHnd);
	if(mydosdevice) DOS_DelDevice(mydosdevice);
};

void CParallel::InstallTimerHandler(TIMER_TickHandler th)
{
	if(TimerHnd==NULL) {
		TimerHnd=th;	
		TIMER_AddTickHandler(th);
	}
}

Bit8u CParallel::getPrinterStatus()
{
	/*	7      not busy
		6      acknowledge
		5      out of paper
		4      selected
		3      I/O error
		2-1    unused
		0      timeout  */
	Bit8u statusreg=Read_SR();

	//LOG_MSG("get printer status: %x",statusreg);
	statusreg^=0x48;
	return statusreg&~0x7;
}

#include "callback.h"

void RunIdleTime(Bitu milliseconds)
{
	Bitu time=SDL_GetTicks()+milliseconds;
	while(SDL_GetTicks()<time)
		CALLBACK_Idle();
}

void CParallel::initialize()
{
	Write_IOSEL(0x55); // output mode
	Write_CON(0xD0); // init on
	Write_PR(0);
	RunIdleTime(10);
	Write_CON(0xD4); // init off
	RunIdleTime(500);
	//LOG_MSG("printer init");
}



CParallel ** parallelPortObjects[3];
class PARPORTS:public Module_base {
public:
	
	PARPORTS (Section * configuration):Module_base (configuration) {
		// put handlers into arrays
		IO_ReadHandler *parallelReadHandlers[] = {
			&PAR1_Read, &PAR2_Read, &PAR3_Read
		};
		IO_WriteHandler *parallelWriteHandlers[] = {
			&PAR1_Write, &PAR2_Write, &PAR3_Write
		};
		TIMER_TickHandler parallelTimerHandlers[] = {
			&PAR1_Update, &PAR2_Update, &PAR3_Update
		};

		// default ports & interrupts
		Bit16u addresses[] = { LPT1_BASE, LPT2_BASE, LPT3_BASE};
		Bit8u defaultirq[] = { 7, 5, 12};
		Bit16u biosParameter[3] = { 0, 0, 0};
		Section_prop *section = static_cast <Section_prop*>(configuration);
		
		const char *configstrings[3] = {
			section->Get_string ("parallel1"),
			section->Get_string ("parallel2"),
			section->Get_string ("parallel3")
		};
		
		parallelPortObjects[0] = &parallel1;
		parallelPortObjects[1] = &parallel2;
		parallelPortObjects[2] = &parallel3;


		// iterate through all 3 lpt ports
		for (Bitu i = 0; i < 3; i++) {
			biosParameter[i] = addresses[i];

			CommandLine* cmd;
			std::string str;
			cmd=new CommandLine(0,configstrings[i]);
			cmd->FindCommand(1,str);

			Bit16u baseAddress = addresses[i];
			Bitu irq = defaultirq[i];
			Bitu realbase = addresses[i];
			biosParameter[i] = baseAddress;

#if C_DIRECTLPT
			if(!str.compare("reallpt")) {

				if(cmd->FindStringBegin("realbase:",str,false)) {
					if(!sscanf(str.c_str(), "%x",&realbase)==1) {
						LOG_MSG("parallel%d: Invalid realbase parameter.",i+1);
						continue;
					}
				}

				CDirectLPT* cdlpt= new CDirectLPT(parallelReadHandlers[i], 
						parallelWriteHandlers[i],parallelTimerHandlers[i],
						baseAddress,defaultirq[i], realbase);
				if(cdlpt->InstallationSuccessful)
					*parallelPortObjects[i]=cdlpt;
				else {
					delete cdlpt;
					parallelPortObjects[i]=0;
				}
			}
			else
#endif
			if(!str.compare("disabled")) {
				*parallelPortObjects[i] = NULL;
				biosParameter[i] = 0;
			} else {
				LOG_MSG ("Invalid type for LPT%d.", i + 1);
				*parallelPortObjects[i] = NULL;
				biosParameter[i] = 0;
			}
		}

		BIOS_SetLPTPorts (biosParameter);
	}

	~PARPORTS () {
		for (Bitu i = 0; i < 3; i++)
			if (parallelPortObjects[i] && *parallelPortObjects[i]) {
				delete *(parallelPortObjects[i]);
				*(parallelPortObjects[i]) = 0;
			}
	}
};

static PARPORTS *testParallelPortsBaseclass;

void PARALLEL_Destroy (Section * sec) {
	delete testParallelPortsBaseclass;
	testParallelPortsBaseclass = NULL;
}

void PARALLEL_Init (Section * sec) {
	// should never happen
	if (testParallelPortsBaseclass) delete testParallelPortsBaseclass;
	testParallelPortsBaseclass = new PARPORTS (sec);
	sec->AddDestroyFunction (&PARALLEL_Destroy, true);
}
